/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 

/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
   Variable parameters: Below are the parameters that you can change 
   without changing the other code in this file (in JavaScript functions)
   xxxxxxxxxxxxxxx     Begin of variable parameters xxxxxxxxxxxxxxxx   */
       
/* raster = This is resolution - an invisible raster that divides the screen into small squares. 
Specifies the maximum number of circles that can be displayed on the screen. 
If this number is higher, it will be possible to show more circles on the screen.
As this number is larger, the circles are smaller and vice versa. 
*/
var raster = 3;

var lc = 4;   // minimum (low) count visible shapes

var hc = 8;   // maximum (high) count visible shapes. Important must  be:  hc <= raster*raster

var w = 1;     // number of item within each level

/* This is basic timer resolution (stimulus exposition time) = 1*time. 
   Units = miliseconds (ms)
   This is also basic application time resolution (minimum)      */
var time = 100;      

/* wait = This is number of basic time cycles before circles (item) exposition. 
    Units: basic timer resolution 
    This is period for draw fixing Cross Cross at centre of  screen (set EVEN number: 2,4,6,8,.. the number to be divisible by 2) 
    How to convert in ms?  Example: if time=500 and wait=8 => 500 * 8 = 4000 ms = 4s 
    Recommended value>= 2  (EVEN number)     */
var wait = 6;         

/* delay = This is delay between stimulus and response. 
  Units: basic timer resolution
  How to convert in ms? [real time delay in miliseconds] = time + (delay * time). 
  Example: if time=700 and delay=5 => [real time delay in miliseconds] =  700 + (5 * 700) = 3500 ms */
var delay = 10;

var monochrome= false;   // if true monochrome (black color circle). false = circles  will be in random Colors 

/* statistics loaded from server database script (wm4.php):
   stat => 0= none (only record ID), 
              stat = 1= count(frequencies) from server databas, 
              stat = 2 = percent from database, 
              stat = 3 = fetch all data from server database     */
var stat = 0;       

var paradigm = 0;   // your optional paradigm  mark (database index. see wm4.php=> create_table)

/*    xxxxxxxxxxxxxxx     End of variable parameters xxxxxxxxxxxxxxxx   */
/*    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx   */


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
"const" parameters => change only if  belonging source code in functions will be changed       
    let and const are not used because of Backward compatibility    
   xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */ 

var itemCnt = w*((hc-lc)+1);    // total item count      
var cor = [];   // correct number of visible shapes within each item
var ans = [];  //  type == string (for google charts) user answers (dependent var)
var xy = [];    //  jagged array [item][] Within each [item] is separate array with grid coordinates [0.index= x; 1.index =y, 2.index = show(1) or hide(0) ]
var timerID = null;   // timerID is the identifier of the repeated action you want to cancel. This ID is returned from setInterval(). 
var item = 0;   // current item
var step = -1;   // global timer counter 
var def_par = {};   // default paramaters


function init() { 
  if (timerID)  {   
    window.clearInterval(timerID);  
  }  
  timerID= setInterval(function () {my_timer()},time); // Returns an intervalID. intervalID is a unique interval ID you can pass to clearInterval()
  def_par = {raster:raster, lc:lc, hc:hc, w:w}; 
  setPar();       
}
 
// show Instruction panel  => form id="data_form"
function showStart(is_repeated)  {      
  document.getElementById("data_form").style.display= "block"; 
  document.getElementById("end").style.display= "none";
  document.getElementById("google-wrapper").style.display="none";
  if (is_repeated) return; // only show Panel with user data   
  if (!validation()) return;        
  document.getElementById("data_form").style.display= "none";
  document.getElementById("start").style.display= "block";
}

function startTest()   {  
  setPar();
  if (!check()) {
     alert("Incorrect parameters: raster to low or hc to high for selected raster. Increase raster or decrease the level of difficulty.");
     return;
  }  
  setCoord(); 
  rndSample();   
  document.getElementById("google-wrapper").style.display="none";
  document.getElementById("end").style.display= "none"; 
  document.getElementById("start").style.display= "none"; 
  item=0;
  step=0;   // activate Timer  
}

// set research (paradigm) parameters
function setPar()  {
  var x = document.getElementsByName("level");
  var lev=0;  
  var h= x.length;    
  for (var i = 0; i <h; i++) {
    if (x[i].checked) {
        lev= x[i].value;
        break;
    }
  }     
  switch(lev) {    
    case "1":        //    light
      raster= 3;  lc=1;  hc=5; w= 2;
      break;
    case "2":         // medium
        raster= 4;  lc=1;  hc=10; w= 1;      
      break;  
    case "3":    // hard
      raster= 4;  lc=6;  hc=15; w= 1;
      break;
    case "4":     // Very hard
      raster= 5;  lc=11;  hc=20; w= 1;
       break;  
    case "5":     // Extreme
      raster= 6;  lc=21;  hc=30; w= 1;
      break;                   
    default:    // default/custom    (case "0")
      raster= def_par["raster"]; lc= def_par["lc"];  hc=def_par["hc"]; w= def_par["w"];
  } 
   itemCnt = w*((hc-lc)+1);     
   cor = [];  
   ans = [];  
   xy = [];        
   document.getElementById("item_cnt").innerHTML= "Number of item: " + itemCnt;
   //console.log("setPar() returns \nlev=" + lev + "\nraster= " + raster + "\nlc=" + lc + "\nhc= " + hc + "\nw= " + w + "\nitemCnt= " + itemCnt);  
}

// this function must be => otherwise there may be an infinite loop 
function check () {
  return (hc <= raster*raster);
}

function setCoord() {
  var strokeWidth = 1;
  var canvas = document.getElementById('canvas');        
  canvas.width= window.innerWidth;       // stretch canvas to parent 
  canvas.height= window.innerHeight;     
  var wmax = canvas.width - raster*strokeWidth,  hmax = canvas.height-raster*strokeWidth; 
  var dx = Math.floor(wmax/(raster+1)) ,  dy= Math.floor(hmax/(raster+1));           
  for (var i=0; i<itemCnt; i++)  {
    xy[i]= [];  
    for (var x = dx; x <= wmax-dx; x += dx) {   
      for (var y = dy; y <= hmax-dy; y += dy) {
        xy[i].push([x, y, 0]);                         
      }
    }
  }
}

function rndSample()  {    
  var tmp=[];      // already choosen items 
  for (var cnt=lc; cnt<=hc; cnt++)  {   //cnt = number of visible shapes within this item
    // choose random item
    for (var j=0; j<w; j++)  {    //   from lf to hc (w-times)
      do
        var item= Math.floor(Math.random()*xy.length);  // random item => xy.length = ok = itemCnt = 1.dim length  
      while (typeof tmp[item] !== 'undefined');    
      tmp[item]=1;
      cor[item]=cnt;  // correct number of visible shapes within each item
      //cnt= (cnt > xy[item].length) ? 1 : cnt;   // redundant?
      // choose cnt random shapes within this random item
      for (var i=0; i<cnt; i++)  {
        do  
          var rnd =  Math.floor(Math.random()*xy[item].length);    
        while (xy[item][rnd][2] == 1);   // check if already choosen
        xy[item][rnd][2]=1;       // this circle will be visible
      }     
    }
  }
}

function getRndInt(min, max) {
    return Math.floor(Math.random() * (max - min + 1) ) + min;
}

function drawCross()  {
  var canvas= document.getElementById("canvas");   //canvas.innerHTML="";
  var x= Math.round(canvas.width/2);
  var y= Math.round(canvas.height/2);
  var d = Math.floor(Math.min(x,y)/4);  // 1/8 min(width, height)  
  drawLine(x-d,y,x+d,y); // horizontal
  drawLine(x,y-d,x,y+d); // vertical 
}

function drawLine(x1,y1,x2,y2)  {
  var ctx= document.getElementById("canvas").getContext('2d'); 
  ctx.beginPath();
  ctx.moveTo(x1,y1);
  ctx.lineTo(x2,y2);
  ctx.lineWidth= 1;
  ctx.strokeStyle="#000000";  // black
  ctx.stroke();     
}

function drawGrid() { 
  var strokeWidth = 1;
  var canvas = document.getElementById("canvas"); 
  //canvas.innerHTML=null;
  var wmax = canvas.width - raster*strokeWidth;
  var hmax = canvas.height - raster*strokeWidth; 
  var dx = Math.floor(wmax/(raster+1)) ,  dy= Math.floor(hmax/(raster+1));      
  var r = Math.floor(Math.min(dx,dy)/4)-2*strokeWidth;       // 1/4 = r within raster rect; rnd offset: +/-r  
  var ctx = document.getElementById("canvas").getContext('2d'); 
  ctx.clearRect(0, 0, canvas.width, canvas.height);      // clear canvas   
  var i= 0;       
  for (var x = dx; x <= wmax-dx; x += dx)  {   
    for (var y = dy; y <= hmax-dy; y += dy)  {                  
      if (xy[item][i][2] === 1)   {
        var xrnd= getRndInt(-r,r); // random x-offset
        var yrnd= getRndInt(-r,r); // random y-offset
        ctx.beginPath();        
        ctx.arc(x+xrnd,y+yrnd,r,0,2*Math.PI);     //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise);     
        ctx.fillStyle = (monochrome==true)  ? ctx.fillStyle =  "black" :   getRandomColor();       //ctx.fillStyle =  (xy[item][i][2] == 0) ?  "black" : "red";    //  getRandomColor();          
        ctx.fill();
      }                 
      i++;           
    }   
  }              
}

function getRandomColor() {
  var letters = '0123456789ABCDEF';
  var color = '#';
  for (var i = 0; i < 6; i++) {
    color += letters[Math.floor(Math.random() * 16)];     // returns a random integer from 0 to 15 
  }
  return color;
}  

function my_timer()  {     
   var canvas= document.getElementById("canvas"); 
   canvas.getContext("2d").clearRect(0, 0, canvas.width, canvas.height);     // clear canvas => Clear Cross
   if (step<0) return;  // nop, idle
   step++; 
   if (step < wait && step % 2 != 0)  {  // Draw Cross only if it is odd number  (1,3,5,..)
     canvas.style.display="block";
     drawCross();   // total time = 3s = 6*500ms = 3xflick
   }
   else if (step == wait+1) {
     drawGrid();     
   }
   else if (step > wait+1+delay) {
      step = -1;         // nop, idle, do none, Timer no operation
      showResponse();    
   }
}

function keyUp(event)  {          
  var e = event ||  window.event;    // event = use the value of event if available or  window.event = if not assume it's IE and use window.event   
  e.preventDefault();            
  if (e.keyCode === 13) {   
    nextItem();         
  }
}

function nextItem()  {
  var node= document.getElementById("answer");
  var tmp = (trim(node.value) == "") ? 0: node.value;  
  if (isInt(tmp) == false || tmp < 0 || tmp > 99) {
    alert("Enter the number of dots (between 0 and 99). If you do not know the answer: enter zero or empty (blank).");
    return;
  }  
  tmp = parseInt(tmp, 10);       
  ans[item]= tmp;     // store validated user answer 
  node.value= "";          // reset user answer (default answer for next item)
  node= document.getElementById("response");
  node.style.display="none";            
  if (item < xy.length - 1) {    // show next task       
    item++;
    step=0;
  }
  else  {               
    document.getElementById("response").style.display="none";
    document.getElementById("end").style.display="block";  
    theEnd();    
  }
}

function trim(str) {
   return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}

// show response div (panel with user answer)
function showResponse()  {
  var node= document.getElementById("canvas");
  node.style.display="none"; 
  var node= document.getElementById("response");
  node.style.display="block"; 
  document.getElementById("answer").focus(); 
}

function theEnd()  {
  var res=0;  
  for (var j=0; j!=cor.length; j++)   {
    if (cor[j] == ans[j]) res++;
  }
  var elem= document.getElementById("score");
  elem.innerHTML="Your score is: " + res + "   from possible: " + cor.length + "   Percentage: " + Math.round(100*res/cor.length) + "%";
  document.getElementById("google-wrapper").style.display="block";
  drawClientData(res);         
  var data= setData(res, stat);
  sendData(data);
}

function drawClientData(res) {    
  //    xxxxxxxxxxx     chart2 => BarChart   xxxxxxxxxxxxxxxxx  
  var data = new google.visualization.DataTable();    // Create and populate the data table. 
  data.addColumn('string', 'Item number');     // 1. col = ans array = must be string  
  data.addColumn('number', 'Answer');      
  data.addColumn('number', 'Correct'); 
  data.addColumn('number', 'Item score');    
  for(var i = 0; i < ans.length; i++)  {  
    var tmp= (ans[i] === cor[i]) ? 1: 0;
    data.addRow( [  (i+1).toString() + ". item",  ans[i],  cor[i], tmp ] );        
  }                    
  options= {
   title: 'Total score = ' + res,
   width: '100%'   
  };    
  document.getElementById("chart2").style.display = "block";   
  new google.visualization.ColumnChart(document.getElementById('chart2')).draw(data, options);     // Create and draw the visualization.    
  //    xxxxxxxxxxxxxx          Frequencies table        xxxxxxxxxxxxxxxxx                             
  options= {      
    'width': '50%',    
    'allowHtml': true, 
    'sort': 'enable',  
    'alternatingRowStyle': true              
  };    // data are same         
  document.getElementById("table").title="Total score=" + res;
  document.getElementById("table").style.display = "block"; 
  new google.visualization.Table(document.getElementById('table')).draw(data, options);     
}
                            
/* check is form-data entry template (client data) are valid. return true = valid or false = invalid                  
  Mozilla:  <input type="number"> elements do not support use of the pattern attribute for making entered values conform to a specific regex pattern.
  see: https://www.w3resource.com/mysql/mysql-data-types.php    
  Be careful: a "validity" object may not have to exist at all and it is not 100% reliable    */
function validation()  {  
  try
  {
    var res= true, first;  // res = function result,  first= first field with error => focus()
    var nodes = document.getElementsByClassName("data");   
    var tmp= nodes.length; 
    for (var i= 0; i<tmp; i++)  { 
      if (!nodes[i].parentNode.parentNode) continue;            
      if (nodes[i].parentNode.parentNode.className === "hide")  continue;
      var my_id= nodes[i].getAttribute("id");   // must have id
      if (!my_id) continue;    
      // build-in browser validation
      var err= (nodes[i].validity.valid) ? false: true;  // this node have error or not
      // wm4 validation
      //document.getElementById("val_" + my_id).innerHTML= null;
      var s= trim(nodes[i].value.toString());
      if (my_id.substr(0,6) === "string")  {     // && !nodes[i].validity.valid
          if (s.length > 50) err= true;
          //nodes[i].value= s.substr(0,50);             // string fields max length (change if you need depending on server DBMS)               
      }     // && !nodes[i].validity.valid 
      else if ((s.length > 0 && my_id.substr(0,3) === "int")  
          && (!isInt(s) || parseInt(s) < -2147483648 || parseInt(s) > 2147483647))  {   // change if you need depending on server DBMS
              err= true;      
      }    // && !nodes[i].validity.valid
      else if ((s.length > 0 && my_id.substr(0,5) === "float")       
          && (!isFloat(s) && !isInt(s)))  {
              err=true;      
      }       
      if (err)  {
          res= false;
          if (!first) first= document.getElementById(my_id);
      }    
      //console.log("node.id = " + nodes[i].getAttribute("id")  + "\n node.value= " + nodes[i].value + "\n getElementById.value= " + document.getElementById(my_id).value + "\n result= " + res + "\n last node= " + err);                                  
      if (document.getElementById("val_"+my_id)) updateVal(my_id, err);  
    }                      
    if (first) first.focus();
  }
  catch (e)  {
    alert("Your browser probably does not support validity object. \n" + "Error Description:\n" + e);
    return true; 
  }  
  return res;   
}

// raise validator alarm => feedback to user
function updateVal(my_id, alarm) {
  var txt="";
  if (my_id.substr(0,3) === "int") {
   txt= "Invalid integer";
  } 
  else if (my_id.substr(0,5) === "float") {
   txt= "Invalid float value";
  } 
  else  {      // else if (my_id.substr(0,6) === "string")
    txt= "Invalid field value";
  }
  document.getElementById("val_"+my_id).innerHTML= (alarm ===true)  ? txt : "";   
}

/*  do basic client data validation again  => a validity object may not have to exist at all and it is not 100% reliable
   Mozilla: <input type="number"> elements do not support use of the pattern attribute for making entered values conform to a specific regex pattern. 
   maxlength is ignored on <input type="number"> by design (Chrome, Firefox)    */
function setData(score, statistics)    {
  var data= {paradigm:paradigm, raster:raster,low:lc, high:hc, within:w,item:itemCnt, timer:time, wait:wait, delay:delay, correct:score, incorrect: itemCnt-score, stat:statistics};      
  var nodes = document.getElementsByClassName("data");   
  var tmp= nodes.length;    
  for (var i= 0; i<tmp; i++)  { 
    if (!nodes[i].parentNode) continue;      
    if (nodes[i].parentNode.parentNode.className === "hide" && nodes[i].getAttribute('data') !== "const" ) continue;
    if (nodes[i].hasAttribute("id") === false) continue;      
    //if (elem.hasAttribute("value") === false) continue;  
    var s= nodes[i].value;  
    if (isNull(s))  continue;
    s= trim(s);    
    if (s.length == 0) continue; 
    var my_id=nodes[i].getAttribute("id"); 
    if (my_id.substr(0,3) === "int")  {     
       s= s.substr(0, 9);    // change if you need depending on server DBMS
      if (!isInt(s)) continue;     
      s=Math.round(parseInt(s));     
    }
    else if (my_id.substr(0,3) === "flo")  {   
        s= s.substr(0, 18);        // change if you need depending on server DBMS
       if (!isFloat(s)) continue;       
    } 
    else  {
        s= s.substr(0, 50);             // string fields max length (change if you need depending on server DBMS)
    }
    data[my_id]= s;                                   
  }    
  data = encodeURIComponent(JSON.stringify(data));
  return data;
}       
     
function sendData(data)  {     
 
  try {        
    var xhr = new XMLHttpRequest();  // Browsers: Mozilla Firefox, Google Chrome, Opera, Microsoft Internet Explorer 7.x-8.x, Apple Safari)
  }
  catch (e) {
    try {            
      var xhr = new ActiveXObject("Microsoft.XMLHTTP");  // Browsers: Microsoft Internet Explorer 5.x-6.x  or backward support => min: Microsoft Internet Explorer 5.x
    }
    catch (e) {
      alert("Your browser does not support XMLHttpRequest() Object. \n" + "Error Description:\n" + e); 
      return;                                
    }
  }
  xhr.onreadystatechange = function() {
      if (this.readyState == 4 ) {
        if (this.status == 200)  {             //alert(this.responseText);
          if (this.responseText && !isEmpty(this.responseText))  { 
            var rsp= this.responseText;        
            if (stat > 0) {
              drawServerData(rsp);
            }
            else if (isInt(rsp)) {
              document.getElementById("server_resp").innerHTML= "Record successfully stored. id=" + rsp;
            }
            else {
              document.getElementById("server_resp").innerHTML= "There was an unexpected error: " + rsp;
            }                             
          }                           
        }
        else {
          alert("Sending data to the web server failed.\nError status code=" + this.status + "\nError Description:\n" + this.responseText );      // tested. ok
        }          
      }
  };
  xhr.open("POST", "php/wm4.php");
  xhr.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
  xhr.send("x=" + data);
} 

//https://developers.google.com/chart/interactive/docs/php_example
function drawServerData(my_json) 
{   
  if (stat === 3)     {
      document.getElementById("table2").innerHTML= my_json;   // response is pure html formated output
      document.getElementById("table2").style.display= "block";
      return;
  }
// *********************    Chart - ColumnChart       ********************* 
  var elem = document.getElementById("chart3");                     
  var data = new google.visualization.DataTable(my_json);  // Create our data table out of JSON data loaded from server.         
  //all_google_types API = [ GeoChart, ScatterChart, Histogram, BarChart, ComboChart, AreaChart, SteppedAreaChart, LineChart, BubbleChart, OrgChart, TreeMap, Table, Timeline, Gauge, CandlestickChart ]         
  var  chart = new google.visualization.ColumnChart(elem);      // Instantiate and draw our chart, passing in some options.
  var my_title= (stat == 1) ? "Frequencies (count) from server database" : "Percent from server database";
  var options = {
    title: my_title, 
    is3D: true,      
    width: '100%'
  };                 
  elem.style.display = "block";        
  chart.draw(data, options);    // set css style before this line    
  // *******************************    TABLE       *********************  
  data = new google.visualization.DataTable(my_json);  // Create our data table out of JSON data loaded from server.              
  options= {      
    'width': '50%',    
    'allowHtml': true, 
    'sort': 'enable',  
    'alternatingRowStyle': true              
  }; 
  document.getElementById("table2").title= my_title;     
  document.getElementById("table2").style.display = "block";    
  new google.visualization.Table(document.getElementById("table2")).draw(data, options);    // data are same  
}

function isEmpty(obj) {
  for(var prop in obj) {
      if(obj.hasOwnProperty(prop))
          return false;
  }
  return true;
}

function isNull(obj)  {
  return  (typeof(obj) === "undefined" || obj === null);
}
                  
function isInt(n) {       
   return (!isNaN(n)) && (parseFloat(n) === parseInt(n, 10)) &&  (n % 1 === 0); 
} 

// change if you need depending on server DBMS. see:  https://www.w3resource.com/mysql/mysql-data-types.php
function isFloat(n) { 
  return !isNaN(parseFloat(n)) && isFinite(n) &&  n > 1E-38 && n < 1E38; 
}  
